//GNU LESSER GENERAL PUBLIC LICENSE - See lgpl.txt
unit SingleInstance;
interface
uses
  Classes,nCommon;

procedure DetachSingleInstance;
function ReSendParams: boolean;
function SendParams(const sApppIdent:AnsiString; params:TWideStringList): boolean;
procedure SetInitialInstanceData;
function ReadParams(Msg: Cardinal): PWideChar;

const
  ParamMessageSep = #13;

implementation

uses
{$ifdef MSWINDOWS}
  Windows,
{$else}
  Types,
{$endif}
  SysUtils,DCPmd5,nStrings,
  Messages, Forms;

const
  PARAMS_SIZE = 1024;

type
  TInstanceDataHeader = record
    InstanceCount: integer;
    hMainform: {$ifdef VisualCLX}QOpenScrollViewH{$else}HWND{$endif};
    idThread: DWORD;
    dir: array[0..260] of WideChar;
    params:array[0..PARAMS_SIZE-1] of WideChar;
  end;
  PInstanceDataHeader = ^TInstanceDataHeader;

var
  hMapping: THandle;
  pHeader: PInstanceDataHeader;
{$ifdef MSWINDOWS}
  WM_PARAM_NOTIFICATION: HWND;
{$endif}
  APP_IDENT : AnsiString;

function MD5DigestToStr(const Digest: array of byte): AnsiString;
var
 i: Integer;
begin
 result:='';
 for i:=0 to 15 do result:=result+IntToHex(Digest[i], 2);
end;

function SendParams(const sApppIdent:AnsiString; params:TWideStringList): boolean;
var
  i: integer;
  str: UnicodeString;
{$ifdef MSWINDOWS}
  Lock: TRTLCriticalSection;
{$endif}
  Digest: array[0..15] of byte;
  Dcp_MD5: TDCP_md5;
begin
  Dcp_MD5:=TDCP_md5.Create(nil);
  Dcp_MD5.Init;
  Dcp_MD5.UpdateStr(sApppIdent);
  Dcp_MD5.Final(Digest);
  Dcp_MD5.Free;
  APP_IDENT := MD5DigestToStr(Digest);
{$ifdef MSWINDOWS}
  try
    InitializeCriticalSection(Lock);
    EnterCriticalSection(Lock);
    hMapping:=OpenFileMapping(
      FILE_MAP_ALL_ACCESS,
      false,
      PChar(APP_IDENT)
      );
    result := (hMapping<>0);
    if not result then
    begin
     hMapping := CreateFileMapping(
        $FFFFFFFF,
        nil,
        PAGE_READWRITE,
        0,
        SizeOf(TInstanceDataHeader)+PARAMS_SIZE,
        PChar(APP_IDENT)
        );
    end;
    pHeader := MapViewOfFile(hMapping,FILE_MAP_ALL_ACCESS,0,0,0);
    Assert(pHeader<>nil);
  finally
    DeleteCriticalSection(Lock);
    LeaveCriticalSection(Lock);
  end;
  if not result then
  begin //not previous instance
    pHeader.InstanceCount := 1;
    pHeader.hMainform := 0;
    exit;
  end;
  if pHeader.InstanceCount<>1 then exit;
  pHeader.InstanceCount:=2;
try
  for i:=0 to 59 do
  begin
    if pHeader.hMainform<>0 then break;
    sleep(50);
  end;
  if pHeader.hMainform=0 then
  begin
    windows.MessageBox(0,PChar('Problem with previous instance Editas in memory'+
            #10'if not appear - try kill it'),'',0);
    exit;
  end;

  //
  {
  gdy za wczesnie wywolany drugi - wychodzi bez wyslania komunikatu
  powinien troch poczekac i zwolnic pamic, a pierwszy, gdy
  nie bdzie juz pamieci dzielonej tez powinien skonczy

  ale co gdy  wywolany bedzie trzeci??
  }
  {
  gdy sie nie doczeka oznacza to ze
  if ReSendParams then halt wykonane
  nie doszlo do Application.Run;
  byc moze koniec podczas tworzenia form
  oraz nie wykolo sie DetachSingleInstance
  czyli program zabity jakims wyjatkiem pozostawil po sobie
  utworzon pamie dzielona
  chyba ze Windows samo zwalnia gdy xzaden program juz nie uzywa (ale chyba nie zzawsze)
  moze rowniez zawisil sie podczas tworzenia jakiejs formy
  }

  WideStrLCopy(pHeader.dir,PWideChar(GetCurrentDirW),260);
  ZeroMemory(@pHeader.params, PARAMS_SIZE);
  str := '';
  for i:=0 to params.Count-1 do
  begin
    str := str + params[i];
    if i<params.Count-1 then str := str + ParamMessageSep;
  end;
  WideStrLCopy(pHeader.params, PWideChar(str), PARAMS_SIZE-1);//ending must be #0
  SetForegroundWindow(pHeader.hMainform);
  SendMessage(pHeader.hMainform, RegisterWindowMessage(PChar(APP_IDENT)), 0, 0)
finally
  pHeader.InstanceCount:=1;
end;
{$else}
  result := false;// yet not supported
{$endif}
end;

function ReSendParams: boolean;
var
  PCmd: PWideChar;
  Params: TWideStringList;
  i: integer;
begin
  PCmd:=GetCommandLineW;
  Params:=TWideStringList.Create;
  SeparateStringsW([' '],[],['"'],PCmd,Params,seMergeSeparators);
  Assert(Params.Count>=1);
  Params.Delete(0); //remove Exe module name from params
  result:=SendParams({$ifdef MSWINDOWS}WideToLower{$endif}(AppPath()),Params);
  Params.Free;
end;

procedure SetInitialInstanceData;
{$ifdef MSWINDOWS}
var
  Lock: TRTLCriticalSection;
{$endif}
begin
{$ifdef MSWINDOWS}
  Application.MainForm.HandleNeeded;
  WM_PARAM_NOTIFICATION := RegisterWindowMessage(PChar(APP_IDENT));
  pHeader.idThread := GetCurrentThreadId;
  pHeader.hMainform := Application.Mainform.Handle;
  ZeroMemory(@pHeader.params, PARAMS_SIZE);
{$else}
// yet not supported
{$endif}
end;

function ReadParams(Msg: Cardinal): PWideChar;
begin
  result:=nil;
{$ifdef MSWINDOWS}
  if Msg<>WM_PARAM_NOTIFICATION then exit;
  if InSendMessage then ReplyMessage(1);
  if pHeader=nil then exit;
  result:=pHeader.params;//another way is SetString and return AnsiString
  SetCurrentDir(pHeader.dir);
{$else}
// yet not supported
{$endif}
end;

procedure DetachSingleInstance;
{$ifdef MSWINDOWS}
var
  Lock: TRTLCriticalSection;
{$endif}
begin
{$ifdef MSWINDOWS}
  try
    InitializeCriticalSection(Lock);
    EnterCriticalSection(Lock);
    if pHeader<>nil then
    begin
      pHeader:=nil;
      UnMapViewOfFile(pHeader);
    end;
    if hMapping<>0 then
    begin
      hMapping:=0;
      CloseHandle(hMapping);
    end;
  finally
    DeleteCriticalSection(Lock);
    LeaveCriticalSection(Lock);
  end;
{$else}
// yet not supported
{$endif}
end;

initialization
finalization
  DetachSingleInstance;
end.